# -*- coding: utf-8 -*-
## PloneSubscription
## Copyright (C)2005 Ingeniweb

## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.

## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

## You should have received a copy of the GNU General Public License
## along with this program; see the file COPYING. If not, write to the
## Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
"""
This module implements generic functions for subscriptions
"""
__version__ = "$Revision: 1.35 $"
# $Source: /cvsroot/ingeniweb/PloneSubscription/SubscriptionTool.py,v $
# $Id: SubscriptionTool.py,v 1.35 2005/11/07 12:46:21 encolpe Exp $
__docformat__ = 'restructuredtext'

# Python imports
import socket
from smtplib import SMTPException

# Zope imports
from Globals import InitializeClass
from AccessControl import ClassSecurityInfo, getSecurityManager
from AccessControl.SecurityManagement import newSecurityManager
from DateTime import DateTime
from Products.MailHost.MailHost import MailHostError

# CMF imports
from Products.CMFCore import CMFCorePermissions
from Products.CMFCore.utils import getToolByName, UniqueObject
from Products.CMFPlone.PropertiesTool import SimpleItemWithProperties

# Archetypes imports
from Products.Archetypes.public import *
from Products.Archetypes.Renderer import renderer

# Products imports
from Products.PloneSubscription.config import *
from Products.PloneSubscription import SubscriptionPermissions
from Products.PloneSubscription.interfaces.ISubscription import ISubscription
from Products.PloneSubscription.interfaces.ISubscriber import ISubscriber
# import from schema
from Products.PloneSubscription.content.schemata import FormAddGroupSubscriptionSchema


_MARKER = 'Marker'

TOOL_META_TYPE = 'SubscriptionTool'

# Subscription tool schema. Only reference field will be editable
schema = BaseFolderSchema.copy() + Schema((
    BooleanField(
        'show_actions',
        searchable=False,
        default=False,
        write_permission=SubscriptionPermissions.ManageSubscriptionContent,
        widget=BooleanWidget(
            visible={ 'view': 'visible', 'edit': 'visible',},
            label="Show actions",
            label_msgid="label_subscription_show_actions",
            description="Show actions permitting to add providers in the site.",
            description_msgid="help_subscription_show_actions",
            i18n_domain="plonesubscription",
            ),
        ),
    ReferenceField(
        'provider',
        multiValued=False,
        required=True,
        searchable=False,
        relationship='subscriptionProviderOf',
        allowed_types=('SubscriptionProvider',),
        read_permission=CMFCorePermissions.View,
        write_permission=SubscriptionPermissions.ManageSubscriptionContent,
        widget=ReferenceWidget(
            destination_types=['SubscriptionProvider',],
            addable=False,
            visible={ 'view': 'visible', 'edit': 'visible',},
            label="Subscription container provider",
            label_msgid="label_subscription_provider",
            description="Folder which provide base subscription functionnality.",
            description_msgid="help_subscription_provider",
            i18n_domain="plonesubscription",
            ),
        ),
    ReferenceField(
        'templates',
        multiValued=False,
        required=True,
        searchable=False,
        relationship='templatesProviderOf',
        allowed_types=('TemplatesProvider',),
        read_permission=CMFCorePermissions.View,
        write_permission=SubscriptionPermissions.ManageSubscriptionContent,
        widget=ReferenceWidget(
            destination_types=['TemplatesProvider',],
            addable=False,
            visible={ 'view': 'visible', 'edit': 'visible',},
            label="Templates provider",
            label_msgid="label_subscription_templates",
            description="Folder which provide base templates.",
            description_msgid="help_subscription_templates",
            i18n_domain="plonesubscription",
            ),
        ),
    IntegerField(
        'periodicity',
        required=True,
        searchable=True,
        default=60,
        vocabulary='getPeriodicityVocabulary',
        enforceVocabulary=True,
        read_permission=CMFCorePermissions.View,
        write_permission=SubscriptionPermissions.ManageSubscriptionContent,
        widget=SelectionWidget(
            label="Mail periodicity",
            label_msgid="label_subscription_periodicity",
            description="Periodicity used to send subscriptions mails.",
            description_msgid="help_subscription_periodicity",
            i18n_domain="plonesubscription",
            ),
        ),
    StringField(
        'base_url',
        required=False,
        searchable=True,
        default='',
        read_permission=CMFCorePermissions.View,
        write_permission=SubscriptionPermissions.ManageSubscriptionContent,
        widget=StringWidget(
            label="Base url",
            label_msgid="label_subscription_base_url",
            description="Site base url.",
            description_msgid="help_subscription_base_url",
            i18n_domain="plonesubscription",
            ),
        ),
    ReferenceField(
        'folderTemplate',
        multiValued=False,
        required=False,
        searchable=False,
        relationship='folderTemplateOf',
        allowed_types=('SubscriptionTemplate',),
        read_permission=CMFCorePermissions.View,
        write_permission=SubscriptionPermissions.ManageSubscriptionContent,
        widget=ReferenceWidget(
            destination_types=['SubscriptionTemplate',],
            addable=False,
            visible={ 'view': 'visible', 'edit': 'visible',},
            label="Folder subscription's template",
            label_msgid="label_subscription_folder_template",
            description="Template for folder subscription.",
            description_msgid="help_subscription_folder_template",
            i18n_domain="plonesubscription",
            ),
        ),
    ReferenceField(
        'keywordsTemplate',
        multiValued=False,
        required=False,
        searchable=False,
        relationship='keywordsTemplateOf',
        allowed_types=('SubscriptionTemplate',),
        read_permission=CMFCorePermissions.View,
        write_permission=SubscriptionPermissions.ManageSubscriptionContent,
        widget=ReferenceWidget(
            destination_types=['SubscriptionTemplate',],
            addable=False,
            visible={ 'view': 'visible', 'edit': 'visible',},
            label="Keywords subscription's template",
            label_msgid="label_subscription_keywords_template",
            description="Template for keywords subscription.",
            description_msgid="help_subscription_keywords_template",
            i18n_domain="plonesubscription",
            ),
        ),
    ),)
schema['id'].widget.visible = { 'view': 'visible', 'edit': 'hidden',}
schema['title'].widget.visible = { 'view': 'visible', 'edit': 'hidden',}
schema['description'].widget.visible = { 'view': 'visible', 'edit': 'hidden',}


class SubscriptionTool(UniqueObject, BaseFolder):
    """A tool to manage placeless subscription"""

    plone_tool = 1
    id = TOOL_ID
    title = "Placeless Subscription tool for Plone"
    portal_type = meta_type = TOOL_META_TYPE
    manage_options = BaseFolder.manage_options

    # AT specific
    archetype_name = "Subscription Tool"
    immediate_view = 'base_view'
    default_view   = 'base_view'
    global_allow = False
    allowed_content_types = ()
    schema = schema
    include_default_actions = 1

    # Tool specific
    security = ClassSecurityInfo()

    actions = (
        {
        'id'            : 'view',
        'name'          : 'View',
        'action'        : 'string:${object_url}/base_view',
        'condition'     : '',
        'permissions'   : (SubscriptionPermissions.ManageSubscriptionContent, ),
        'category'      : 'object',
        'visible'       : 1,
        },
        {
        'id'            : 'edit',
        'name'          : 'Edit',
        'action'        : 'string:${object_url}/base_edit',
        'condition'     : '',
        'permissions'   : (SubscriptionPermissions.ManageSubscriptionContent, ),
        'category'      : 'object',
        'visible'       : 1,
        },
        {
        'id'            : 'manage_subscribers',
        'name'          : 'action_subscription_manage_subscribers',
        'action'        : 'python:portal.portal_subscription.getSubscriptionProvider().absolute_url()',
        'condition'     : '',
        'permissions'   : (SubscriptionPermissions.ManageSubscriptionContent, ),
        'category'      : 'object',
        'visible'       : 1,
        },
        {
        'id'            : 'manage_mail_templates',
        'name'          : 'action_subscription_manage_mail_templates',
        'action'        : 'python:portal.portal_subscription.getTemplates().absolute_url()',
        'condition'     : '',
        'permissions'   : (SubscriptionPermissions.ManageSubscriptionContent, ),
        'category'      : 'object',
        'visible'       : 1,
        },
    )

    # initialization

    def __init__(self):
        BaseFolder.__init__(self, self.id)

    # accessor and mutator overload

    security.declarePublic('showActions')
    def showActions(self):
        """give public accessor for show_actions"""
        return self.getShow_actions()

    security.declareProtected(SubscriptionPermissions.ManageSubscriptionContent, 'setPeriodicity')
    def setPeriodicity(self, periodicity, **kw):
        """Synchronise field data with subscription_task"""
        self.getField('periodicity').set(self, periodicity)

        mtool = getToolByName(self, 'portal_maintenance')
        subscription_task = getattr(mtool.tasks, subscription_task_id)

        if periodicity == '60':
            subscription_task.scheduled_month = '*'
            subscription_task.scheduled_day_of_month = '*'
            #subscription_task.scheduled_day_of_week = '*'
            subscription_task.scheduled_hour = '*'
            subscription_task.scheduled_minute = '1'
        elif periodicity == '1440':
            subscription_task.scheduled_month = '*'
            subscription_task.scheduled_day_of_month = '*'
            #subscription_task.scheduled_day_of_week = '*'
            subscription_task.scheduled_hour = '1'
            subscription_task.scheduled_minute = '1'
        elif periodicity == '10080':
            subscription_task.scheduled_month = '*'
            subscription_task.scheduled_day_of_month = '*'
            #subscription_task.scheduled_day_of_week = '*'
            subscription_task.scheduled_hour = '1'
            subscription_task.scheduled_minute = '1'
        elif periodicity == '43200':
            subscription_task.scheduled_month = '*'
            subscription_task.scheduled_day_of_month = '1'
            #subscription_task.scheduled_day_of_week = '1'
            subscription_task.scheduled_hour = '1'
            subscription_task.scheduled_minute = '1'
        elif periodicity == '525600':
            subscription_task.scheduled_month = '1'
            subscription_task.scheduled_day_of_month = '1'
            #subscription_task.scheduled_day_of_week = '1'
            subscription_task.scheduled_hour = '1'
            subscription_task.scheduled_minute = '0'

    # methods

    security.declarePublic('getAllPortalTypes')
    def getMemberId(self, default=_MARKER):
        """Return authenticated member id"""
        if default is _MARKER:
            default = None
        mtool = getToolByName(self, 'portal_membership')
        member = mtool.getAuthenticatedMember()

        if not member:
            return default
        else:
            return member.getId()

    security.declarePublic('getMemberSubscriptionUids')
    def getMemberSubscriptionUids(self):
        """Return the list of member's Subscription UID"""
        return [subscription.UID() for subscription in self.getSubscriptions()]

    security.declarePublic('listSubscriberByType')
    def listSubscriberByType(self, portal_type=''):
        if not portal_type:
            return ()
        catalog = getToolByName(self, 'portal_catalog')
        return tuple([brain.getObject() for brain in catalog.searchResults(portal_type=portal_type)])

    security.declareProtected(SubscriptionPermissions.ManageSubscriptionContent, 'listSubscriberTypeSubscriptions')
    def listSubscriberTypeSubscriptions(self, portal_type=''):
        """Return all group subscription without filter"""
        subscribers = self.listSubscriberByType(portal_type=portal_type)
        subscriptions = []
        for subscriber in subscribers:
            subscriptions += subscriber.objectValues()
        return tuple(subscriptions)

    security.declarePublic('hasSubscription')
    def hasSubscription(self):
        """Return True if the user has one subscription

        It's not possible to use subscription catalog here since catalog catch
        REQUEST even if no argument or REQUEST=None is given in argument.
        """
        #if len(self.getSubscriptions) > 0:
        #    return True
        #else:
        #    return False
        return True

    security.declarePublic('getSubscriptionProvider')
    def getSubscriptionProvider(self, default=_MARKER):
        """Return the folder where member's subscription folder are stored by default"""
        if default is _MARKER:
            default = None
        provider = self.getProvider()
        if provider is None:
            return default
        else:
            return provider

    security.declarePublic('getSubscriber')
    def getSubscriber(self, id='', default=_MARKER):
        """Return subscriber where subscriptions are stored by default"""
        if default is _MARKER:
            default = None
        provider = self.getSubscriptionProvider()
        # if no id is given, we assume that's the authenticated member's id
        if not id:
            id = self.getMemberId()

        if not provider:
            return default

        subscriber = getattr(provider, id, _MARKER)
        if subscriber is _MARKER:
            return default
        else:
            return subscriber

    security.declarePublic('getSubscriberUrl')
    def getSubscriberUrl(self, id=''):
        """Return the rpath where an user can see his Subscription"""
        subscriber = self.getSubscriber(id, None)
        if subscriber:
            return subscriber.absolute_url()
        else:
            return getToolByName(self, 'portal_url')

    security.declareProtected(SubscriptionPermissions.AddSubscriptionContent, 'setSubscription')
    def setSubscription(self, id='', ptype='', subscriber_uid='', **kw):
        """Create or modify a subscription for an user

        If you have a existing ptype implementing ISubscription we try to catch
        or create the object then we modify it with keywords argument.

        Keywords have to contains rpath and title key.
        """
        # Do not create anything if the portal type doesn't exist
        types_tool = getToolByName(self, 'portal_types')
        uid_catalog = getToolByName(self, 'uid_catalog')
        if ptype not in types_tool.objectIds():
            return None

        if subscriber_uid:
            subscriber = uid_catalog(UID=subscriber_uid)[0].getObject()
        else:
            subscriber = self.getSubscriber()

        if subscriber is None:
            self.addSubscriber()
            subscriber = self.getSubscriber()

        if id == '':
            id = subscriber.generateUniqueId(ptype)
        # If subscription exist, we will update it
        subscription = getattr(subscriber, id, None)
        if subscription is None:
            try:
                subscriber.invokeFactory(ptype, id)
            except ValueError:
                return None
            subscription = getattr(subscriber, id, None)

        if not ISubscription.isImplementedBy(subscription):
            return None

        subscription.edit(**kw)
        if kw.has_key('title'):
            subscription.setTitle(kw['title'])

        return subscription

    security.declarePublic('getSubscriptions')
    def getSubscriptions(self):
        """Return Subscription stored in a SubscriptionFolder list in a dict"""
        return tuple(self.getCatalog().searchResults())

    security.declareProtected(SubscriptionPermissions.AddSubscriptionContent, 'addSubscriber')
    def addSubscriber(self):
        """Create a member subscriber folder only"""
        provider = self.getSubscriptionProvider()
        member_id = self.getMemberId()

        provider.invokeFactory('UserSubscriber', id=member_id)

    security.declarePublic('getCatalog')
    def getCatalog(self):
        """Return provider catalog"""
        provider = self.getSubscriptionProvider()
        if provider is None:
            return None
        else:
            return provider.getCatalog()

    security.declarePublic('getSubcriberFor')
    def getSubcribersFor(self, rpath=''):
        """Return all subscriber for a subscription """
        catalog = self.getCatalog()
        results = catalog.searchResults(rpath=rpath, sort='getLast_send')

    security.declarePublic('getPeriodictyVocabulary')
    def getPeriodicityVocabulary(self):
        """Return strict periodicity vocabulary

        Use minutes for key to help DateTime operations
        """
        return Vocabulary(DisplayList((
            ('60', "Hourly", 'periodicity_hourly'),
            ('1440', "Daily", 'periodicity_daily'),
            # Not implemented in PloneMaintenance for the moment
            #('10080', "Weekly", 'periodicity_weekly'),
            ('43200', "Monthly", 'periodicity_monthly'),
            ('525600', "Yearly", 'periodicity_yearly'),
            )), self, 'plonesubscription')

    security.declarePublic('getGroupSubscriberVocabulary')
    def getGroupSubscriberVocabulary(self):
        """ Return a Vocabulary of created GroupSubscriber for adding group
            subscription interface
        """
        return Vocabulary(DisplayList(
            [(ob.UID(), ob.title_or_id()) for ob in self.listSubscriberByType(portal_type='GroupSubscriber')]
            ), self, 'plonesubscription')

    security.declareProtected(SubscriptionPermissions.ViewSubscriptionContent, 'getSubscriptionTemplate')
    def getSubscriptionTemplate(self, subscription=None):
        """get the template to use with a subscription"""
        ptype = subscription.portal_type

        if ptype == 'FolderSubscription' and self.getFolderTemplate() is not None:
            template = self.getFolderTemplate().getMail_body()

        elif ptype == 'KeywordsSubscription' and self.getKeywordsTemplate() is not None:
            template = self.getKeywordsTemplate().getMail_body()

        else:
            template = """
Your Subscription to "%(subscription_title)s" push you new data:

%(content)s
"""

        return template


    security.declareProtected(SubscriptionPermissions.ViewSubscriptionContent, 'getUpdatedSubscriptions')
    def getUpdatedSubscriptions(self, date):
        """Get all updated subscription since a date"""
        return tuple(self.getCatalog().searchResults(
            getLast_Send={
                'query': [date - float(self.getPeriodicity())/1440.0,],
                'range': 'max'
                },
            ))

    security.declareProtected(SubscriptionPermissions.ManageSubscriptionContent, 'mailing')
    def mailing(self):
        """get all information and send mails"""
        portal = getToolByName(self, 'portal_url').getPortalObject()
        portal_path = getToolByName(self, 'portal_url').getPortalPath()
        mtool = getToolByName(self, 'portal_membership')
        ptool = getToolByName(self, 'portal_properties')
        now = DateTime()

        # get mail headers data
        date = now.rfc822()
        email_from_address = portal.getProperty('email_from_address', 'admin@plone-subscription.com')
        charset = ptool.site_properties.getProperty('default_charset', 'utf-8')
        headers_infos = {
            'date': date,
            'mail_from': email_from_address,
            'mail_to': '',
            'subject': '',
            'charset': charset,
            }

        current_user = getSecurityManager().getUser()
        subscriptions = self.getUpdatedSubscriptions(now)
        portal_path_length = len(portal_path)
        for sub_brain in subscriptions:
            subscription = sub_brain.getObject()

            # Take care of subscription owner rights
            users = subscription.getSubscriber().getUsers(subscription.id)
            for user in users:
                newSecurityManager(None, user)

                brains = subscription.afterQueryFilters(subscription.getQueryBrains())

                # Restore current user permission
                newSecurityManager(None, current_user)

                template = self.getSubscriptionTemplate(subscription)
                if brains in ([], ()):
                    continue

                content = ''
                for brain in brains:
                    content += "    " + str(brain.Title or brain.getId) + " :\n"
                    base_url = self.getBase_url()
                    if base_url == '':
                        content += str(brain.getURL()) + "\n\n"
                    else:
                        content += base_url + str(brain.getURL(relative=1)[portal_path_length:]) + "\n\n"

                body =  template % {
                    'subscription_title': subscription.TitleOrId(),
                    'content': content,
                    }

                member = mtool.getMemberById(user.getId())
                # Check if member has subscribed and if he has an email
                if member is None:
                    # perhaps an user from Zope root acl_users
                    continue
                member_email = member.getProperty('email', '')
                if member_email == '':
                    continue

                headers_infos['mail_to'] = member_email
                headers_infos['subject'] = subscription.Title()

                if self.sendMail(self.getHeaders(headers_infos) + "\n\n" + body):
                    subscription.setLast_send(now)

        Log(LOG_DEBUG, "Mailing done")
        return 'mailing done'


    security.declareProtected(SubscriptionPermissions.ManageSubscriptionContent, 'getHeaders')
    def getHeaders(self, headers_infos={}):
        """ return mail headers

        take a dictionnary with five item:
        - date
        - mail_from
        - mail_to
        - subject
        - charset
        """

        headers = """Date: %(date)s
From: %(mail_from)s
To: %(mail_to)s
Subject: %(subject)s
Content-Type: text/plain; charset=%(charset)s
"""
        return headers % headers_infos

    security.declarePrivate('sendMail')
    def sendMail(self, messageText, mto=None, mfrom=None, subject=None, encode=None):
        """Send a preformatted mail"""

        Log(LOG_DEBUG, "SubscriptionTool.sendMail : mail body\n\n %s" % messageText)
        try:
            portal = getToolByName(self, 'portal_url').getPortalObject()
            portal.MailHost.send(messageText, mto, mfrom, subject, encode)
            return True
        except (socket.error, MailHostError, SMTPException), error_msg:
            Log(LOG_INFO, "SubscriptionTool.sendMail",
                "Error while sending mail:\n %s" % error_msg)
            return False

    # rendering utils

    security.declarePublic('renderEmptyWidget')
    def renderEmptyWidget(self, instance, field, mode, **kwargs):
        """Render empty widget from field"""

        accessor = field.getAccessor(instance)
        widget = field.widget
        field_name = field.getName()
        return renderer.render(field_name=field_name, mode=mode, widget=widget,
                               instance=instance, field=field, accessor=accessor,
                               **kwargs)

    security.declarePublic('getFormAddGroupedSubscriptionFields')
    def getFormAddGroupedSubscriptionFields(self):
        """ Return adding grouped subscription form schema fields """
        return FormAddGroupSubscriptionSchema.fields()

registerType(SubscriptionTool, PROJECTNAME)

# These methods would be useful with Zope 2.8 to get types implementing some
# services interfaces.

def getInterfaceImplementedBy(interface):
    """Return a list of portal_types implementing an interface"""
    return tuple([type['name'] for type in listTypes()
                  if interface.isImplementedByInstancesOf(type['klass'])])

def getSubscriberAllowedContentTypes():
    """Return portal_types implementing ISubscription interface"""
    return getInterfaceImplementedBy(ISubscription)

def getProviderAllowedContentTypes():
    """Return portal_types implementing ISubscriber interface"""
    return getInterfaceImplementedBy(ISubscriber)
